package net.locusworks.argparser;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import net.locusworks.argparser.annotations.Parameter;
import net.locusworks.argparser.interfaces.IParameterConverter;

public class ArgParser {

  private Class<?> argClass;
  private List<Field> fields;
  private String programName;

  private ArgParser() { }

  public static ArgParser newBuilder() {
    return new ArgParser();
  }

  public ArgParser withArgClass(Class<?> argClass) {
    this.argClass = argClass;
    recollectFields();
    return this;
  }

  private void recollectFields() {
    this.fields = Stream.of(argClass.getDeclaredFields())
        .filter(f -> f.isAnnotationPresent(Parameter.class))
        .map(f -> {
          f.setAccessible(true);
          return f;
        })
        .collect(Collectors.toList());
    
    if (fields == null || fields.isEmpty()) throw new IllegalArgumentException("No fields with @Parameter defined");
    
    fields.sort(new Comparator<Field>() {

      @Override
      public int compare(Field o1, Field o2) {
        Integer o1Order = o1.getDeclaredAnnotation(Parameter.class).order();
        Integer o2Order = o2.getDeclaredAnnotation(Parameter.class).order();
        return o1Order.compareTo(o2Order);
      }
    });
    
    boolean positional = true;
    for (Field f : fields) {
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      positional &= param.positional();
      if (!positional && param.positional()) throw new IllegalArgumentException("All positional values have to be first in list");
    }
  }

  public ArgParser withProgramName(String name) {
    this.programName = name;
    return this;
  }

  public void help() {
    recollectFields();
    String name = programName == null || programName.isEmpty() ? "program" : programName; 
    StringBuilder sb = new StringBuilder();
    sb.append(String.format("Usage: java -jar %s ", name));
    
    for(Field f: fields) {
      Parameter param = f.getAnnotation(Parameter.class);
      
      if (param.positional()) {
        sb.append(f.getName().toUpperCase());
        sb.append(" ");
        continue;
      }
      
      if(!param.required()) {
        sb.append("[");
      }
      
      sb.append(String.format("%s %s", param.names()[0], f.getName().toUpperCase()));
      
      if(!param.required()) {
        sb.append("]");
      }
      
      sb.append(" ");
    }
    
    sb.append("\n");
    
    int longestLength = 0;
    
    for (Field f: fields) {
      Parameter param = f.getAnnotation(Parameter.class);
      if (param.positional()) {
        if (f.getName().length() > longestLength) {
          longestLength = f.getName().length();
        }
        continue;
      }
      String options  = Stream.of(param.names()).collect(Collectors.joining(" | "));
      
      if (options.length() > longestLength) {
        longestLength = options.length();
      }
    }
    
    
    for(Field f: fields) {
      Parameter param = f.getAnnotation(Parameter.class);
      
      if (param.positional()) {
        sb.append(String.format("\t%-" + longestLength +"s : %s%n", f.getName().toUpperCase(), param.description()));
        continue;
      }
      
      String options  = Stream.of(param.names()).collect(Collectors.joining(" | "));
      
      sb.append(String.format("\t%-" + longestLength +"s : %s %s%n", options, param.description(), param.required() ? "[required]": ""));
    }
    
    System.out.println(sb.toString());
  }

  public <T> T parse(String[] args) {
    try {
      return parseItem(args);
    } catch (Exception ex) {
      System.err.println("An exception occured: " + ex.getMessage());
      help();
      
      ex.setStackTrace(new StackTraceElement[0]);
      if (ex instanceof IllegalArgumentException) throw (IllegalArgumentException) ex;
      IllegalArgumentException e = new IllegalArgumentException(ex.getMessage());
      e.setStackTrace(new StackTraceElement[0]);
      throw e;
    }
  }

  @SuppressWarnings("unchecked")
  private <T> T parseItem(String[] args) throws InstantiationException, IllegalAccessException {
    if (argClass == null) throw new IllegalArgumentException("Argument class was not specified");
    
    List<String> argList = new ArrayList<>(Arrays.asList(args));

    Optional<String> helpTest = argList.stream().filter(s -> s.trim().equals("-h") || s.trim().equals("--help")).findFirst();
    if (helpTest.isPresent()) {
      help();
      System.exit(0);
    }
    
    Object obj = argClass.newInstance();
    recollectFields();
    
    List<Field> positional = new ArrayList<>();
    for(Iterator<Field> iter = fields.iterator(); iter.hasNext();) {
      Field f = iter.next();
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      if (param.positional()) {
        positional.add(f);
        iter.remove();
      }
    }
    
    for(Iterator<String> iter = argList.iterator(); iter.hasNext() && !positional.isEmpty();) {
      String value = iter.next();
      Field f = positional.remove(0);
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      IParameterConverter<?> converter = param.converter().newInstance();
      f.set(obj, converter.convert(value));
      iter.remove();
    }
    
    Map<Field, List<String>> paramMap = new LinkedHashMap<>();
    List<String> params = null;
    for(Iterator<String> iter = argList.iterator(); iter.hasNext();) {
      String value = iter.next();
      Field f = findField(value);
      if (f != null) {
        if (!paramMap.containsKey(f)) {
          paramMap.put(f, new ArrayList<>());
        }
        params = paramMap.get(f);
      } else {
        params.add(value);
      }
      iter.remove();
    }
    
    for(Entry<Field, List<String>> entry : paramMap.entrySet()) {
      Field f = entry.getKey();
      List<String> values = entry.getValue();
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      if (param.variableArity() && !f.getType().isAssignableFrom(List.class))
        throw new IllegalArgumentException(String.format("Field %s is not of type list for the params", f.getName()));
      IParameterConverter<?> converter = param.converter().newInstance();
      List<Object> objs = values.stream().map(s -> converter.convert(s)).collect(Collectors.toList());
      if (param.variableArity()) {
        f.set(obj, getGenericList(f.getType(), objs));
      } else {
        f.set(obj, objs.get(0));
      }
    }
    
    for(Field f : fields) {
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      if (param.required() && f.get(obj) == null) {
        throw new IllegalArgumentException(String.format("Field %s is a required field. Value was empty", f.getName()));
      }
    }
    
    return (T)obj;
  }

  private Field findField(String value) {
    for (Field f : fields) {
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      for (String s : Arrays.asList(param.names())) {
        if (s.equals(value)) return f;
      }
    }
    return null;
  }
  
  @SuppressWarnings("unchecked")
  private <Type> List<Type> getGenericList(Class<Type> type, List<Object> items) {
    List<Type> l = new ArrayList<>();
    for (Object o : items) {
      l.add((Type)o);
    }
    return l;
  }
}
